@@ -86,14 +86,17 @@ def refine_r(
     n,
     lam,
     d_low,
+    r_min,
     r_max,
     custom_cost,
-    verbose=False,
+    verbose=0,
     ):
 
+    lb = np.zeros(r0.size)
+    lb[0] = r_min
     ub = np.inf*np.ones(r0.size)
     ub[0] = r_max
-    bnd = Bounds(np.zeros(r0.size), ub)
+    bnd = Bounds(lb, ub)
 
     A = np.zeros((r0.size, r0.size))
     for l in range(r0.size):
@@ -118,7 +121,7 @@ def refine_r(
             constraints=constr,
             bounds=bnd,
             options={
-                'verbose': 2 if verbose and comm.rank==0 else 0,
+                'verbose': 2 if verbose >= 3 and comm.rank==0 else 0,
                 'gtol': 1e-8,
                 'xtol': 1e-8,
                 'maxiter': 1000,
@@ -146,7 +149,10 @@ def refine_r(
             ml.eta_tilde,
         )       
     
-    except:
+    except Exception as e:
+        if verbose >= 1:
+            print(e, flush=True)
+            
         r_new = np.nan*r0
         cost = np.nan
         Q_sca = None
@@ -226,7 +232,7 @@ def insert_needle(
     d_low,
     custom_cost,
     lmax=None,
-    verbose=False,
+    verbose=0,
     ):
     """ mat_dict: database of all materials in the multilayer stack
         mat_needle: list of materials that can be inserted as needles (array of strings)
@@ -364,7 +370,7 @@ def insert_needle(
             intervals = np.size(loc_temp)
             loc_copy = loc_temp.copy()
             cnt = 0
-            if verbose and comm.rank==0:
+            if verbose >= 2 and comm.rank==0:
                 print('Minima Locations Identified    : ', end='', flush=True)
                 for ind in range(intervals):
                     print(str(np.round(loc_copy[ind], 2)) + ' | ', end='', flush=True)
@@ -418,7 +424,7 @@ def insert_needle(
                         ),
                         bounds=loc_copy[i:i+2],
                         method='bounded',
-                        options={'disp': 3 if verbose and comm.rank==0 else 0},
+                        options={'disp': 3 if verbose >= 3 and comm.rank==0 else 0},
                     )
                                              
                     nfev += result.nfev
@@ -461,10 +467,11 @@ def deep_search(
     n,
     ban_needle,
     d_low,
+    r_min,
     r_max,
     custom_cost,
     lmax=None,
-    verbose=False,
+    verbose=0,
     ):
                 
     MF_deep = np.zeros(mat_needle.size)
@@ -514,6 +521,7 @@ def deep_search(
                     n_new,
                     ml_init.lam,
                     d_low,
+                    r_min,
                     r_max,
                     custom_cost,
                     verbose=verbose,
@@ -531,6 +539,8 @@ def deep_search(
             indMF_deep[m] = np.nanargmin(MF_deep_temp)
 
     if np.sum(np.isnan(MF_deep)) == MF_deep.size:
+        needle_status = 0
+    
         ml_init.update(r, n, lmax=lmax)
         Q_sca, Q_abs, Q_ext, p, diff_CS, t_El, t_Ml, Q_sca_mpE, Q_sca_mpM,\
             S1_mpE, S1_mpM, S2_mpE, S2_mpM = tmm.efficiencies(
@@ -572,8 +582,10 @@ def deep_search(
                     ml_init.eta_tilde,
                 )
 
-        return n, r, ban_needle, mat_profile, Q_sca, Q_abs, Q_ext, p, diff_CS
+        return n, r, ban_needle, mat_profile, Q_sca, Q_abs, Q_ext, p, diff_CS, needle_status
     else:
+        needle_status = 1
+    
         mat_final = np.nanargmin(MF_deep)
         n_out = n_out[mat_final,indMF_deep[mat_final]]
         r_out = r_out[mat_final,indMF_deep[mat_final]]
@@ -618,7 +630,7 @@ def deep_search(
                     ml_init.eta_tilde,
                 )
         
-        return n_out, r_out, ban_needle_out[mat_final,indMF_deep[mat_final]], mat_profile_out[mat_final,indMF_deep[mat_final]], Q_sca, Q_abs, Q_ext, p, diff_CS
+        return n_out, r_out, ban_needle_out[mat_final,indMF_deep[mat_final]], mat_profile_out[mat_final,indMF_deep[mat_final]], Q_sca, Q_abs, Q_ext, p, diff_CS, needle_status
 
 def run_needle(
     index,
@@ -636,12 +648,20 @@ def run_needle(
     theta_plot,
     phi_plot,
     d_low,
+    r_min,
     r_max,
     max_layers,
     custom_cost,
     lmax=None,
-    verbose=False,
+    verbose=0,
     ):
+    
+    if verbose >= 1 and comm.rank == 0:
+        print('\n### Optimization Start', flush=True)
+        print('Initial Design: ', end='', flush=True)
+        for ind in range(r.size):
+            print(str(np.round(r[ind], 1)) + ' | ', end='', flush=True)
+        print('', flush=True)
                
     ml_init = multilayer(lam_cost, theta_cost, phi_cost)
     ml_init.update(r, n, lmax=lmax)
@@ -655,13 +675,14 @@ def run_needle(
             n,
             lam_cost,
             d_low,
+            r_min,
             r_max,
             custom_cost,
             verbose=verbose,
         )
     
-    if verbose and comm.rank == 0:
-        print('Current Design: ', end='', flush=True)
+    if verbose >= 1 and comm.rank == 0:
+        print('Iteration ' + str(iteration) + ' Design: ', end='', flush=True)
         for ind in range(r_new.size):
             print(str(np.round(r_new[ind], 1)) + ' | ', end='', flush=True)
         print('', flush=True)
@@ -691,17 +712,17 @@ def run_needle(
             verbose=verbose,
         )
         
-        if verbose and comm.rank == 0:
+        if verbose >= 2 and comm.rank == 0:
             if needle_status:
-                print('Needle Insertion Location Found', flush=True)
+                print('\nNeedle Insertion Location Found', flush=True)
             else:
-                print('Needle Insertion Not Possible', flush=True)
+                print('\nNeedle Insertion Not Possible', flush=True)
         
         if needle_status == 0:
             break
         
         n_new, r_new, ban_needle_new, mat_profile_new,\
-            Q_sca_new, Q_abs_new, Q_ext_new, p_new, diff_CS_new = deep_search(
+            Q_sca_new, Q_abs_new, Q_ext_new, p_new, diff_CS_new, needle_status = deep_search(
                 index,
                 ml_init,
                 mat_needle,
@@ -713,14 +734,24 @@ def run_needle(
                 n_new,
                 ban_needle_new,
                 d_low,
+                r_min,
                 r_max,
                 custom_cost,
                 lmax=lmax,
                 verbose=verbose,
             )
         
-        if verbose and comm.rank == 0:
-            print('Current Design: ', end='', flush=True)
+        if verbose >= 2 and comm.rank == 0:
+            if needle_status:
+                print('Deep Search Done', flush=True)
+            else:
+                print('All Needle Insertion Locations Invalid', flush=True)
+        
+        if needle_status == 0:
+            break
+        
+        if verbose >= 1 and comm.rank == 0:
+            print('\nIteration ' + str(iteration) + ' Design: ', end='', flush=True)
             for ind in range(r_new.size):
                 print(str(np.round(r_new[ind], 1)) + ' | ', end='', flush=True)
             print('', flush=True)
@@ -776,6 +807,7 @@ def run_needle(
                 n_fin,
                 lam_cost,
                 d_low,
+                r_min,
                 r_max,
                 custom_cost,
                 verbose=verbose,
@@ -798,6 +830,13 @@ def run_needle(
     Q_sca_fin, Q_abs_fin, Q_ext_fin, p_fin, diff_CS_fin, t_El, t_Ml, Q_sca_mpE, Q_sca_mpM,\
                 S1_mpE, S1_mpM, S2_mpE, S2_mpM = sim.simulate(lam_plot, theta_plot, phi_plot, r_fin, n_fin)
     
+    if verbose >= 1 and comm.rank == 0:
+        print('\n### Optimization Done', flush=True)
+        print('Final Design: ', end='', flush=True)
+        for ind in range(r_fin.size):
+            print(str(np.round(r_fin[ind], 1)) + ' | ', end='', flush=True)
+        print('', flush=True)
+    
     return r_fin, n_fin, Q_sca_fin, Q_abs_fin, Q_ext_fin, p_fin, diff_CS_fin, cost
 
 def radius_sweep(
@@ -818,6 +857,7 @@ def radius_sweep(
     custom_cost,
     mat_data_dir=None,
     lmax=None,
+    N_final=1,
     verbose=False,
     ):
 
@@ -877,6 +917,7 @@ def radius_sweep(
             theta_plot,
             phi_plot,
             d_low,
+            r_min,
             r_max,
             max_layers,
             custom_cost,
@@ -949,6 +990,7 @@ def radius_sweep(
                 theta_plot,
                 phi_plot,
                 d_low,
+                r_min,
                 r_max,
                 max_layers,
                 custom_cost,
@@ -1038,14 +1080,14 @@ def radius_sweep(
         # Save Best Design
         cost_sort = np.argsort(cost)
         cost = cost[cost_sort]
-        r_save = r_save[cost_sort,:][0,:]
-        n_save = n_save[cost_sort,:,:][0,:,:]
-        Q_sca = Q_sca[cost_sort,:][0,:]
-        Q_abs = Q_abs[cost_sort,:][0,:]
-        Q_ext = Q_ext[cost_sort,:][0,:]
-        p = p[cost_sort,:,:,:][0,:,:,:]
-        diff_CS = diff_CS[cost_sort,:,:,:][0,:,:,:]
-        N_layer = N_layer[cost_sort][0]
+        r_save = r_save[cost_sort,:][:N_final,:]
+        n_save = n_save[cost_sort,:,:][:N_final,:,:]
+        Q_sca = Q_sca[cost_sort,:][:N_final,:]
+        Q_abs = Q_abs[cost_sort,:][:N_final,:]
+        Q_ext = Q_ext[cost_sort,:][:N_final,:]
+        p = p[cost_sort,:,:,:][:N_final,:,:,:]
+        diff_CS = diff_CS[cost_sort,:,:,:][:N_final,:,:,:]
+        N_layer = N_layer[cost_sort][:N_final]
             
         np.savez(
             output_filename,
